""" Generators for basis evaluations """

from dune.perftool.generation import (backend,
                                      class_member,
                                      generator_factory,
                                      get_backend,
                                      include_file,
                                      instruction,
                                      kernel_cached,
                                      preamble,
                                      temporary_variable,
                                      )
from dune.perftool.pdelab.spaces import (lfs_child,
                                         lfs_iname,
                                         lfs_inames,
                                         name_leaf_lfs,
                                         name_lfs,
                                         name_lfs_bound,
                                         type_gfs,
                                         )
from dune.perftool.pdelab.geometry import (dimension_iname,
                                           name_dimension,
                                           name_jacobian_inverse_transposed,
                                           to_cell_coordinates,
                                           )
from dune.perftool.pdelab.localoperator import (lop_template_ansatz_gfs,
                                                lop_template_test_gfs,
                                                )
from dune.perftool.tools import (get_pymbolic_basename,
                                 get_pymbolic_indices,
                                 )
from dune.perftool.pdelab.driver import FEM_name_mangling
from dune.perftool.pdelab.restriction import restricted_name
from pymbolic.primitives import Product, Subscript, Variable
from loopy import Reduction


def type_localbasis_cache(element):
    return "LocalBasisCacheWithoutReferences<{}>".format(type_gfs(element))


@class_member(classtag="operator")
def define_localbasis_cache(element, name):
    include_file("dune/perftool/localbasiscache.hh", filetag="operatorfile")
    t = type_localbasis_cache(element)
    return "{} {};".format(t, name)


def name_localbasis_cache(element):
    name = "cache_{}".format(FEM_name_mangling(element))
    define_localbasis_cache(element, name)
    return name


def declare_cache_temporary(element, restriction, which):
    t_cache = type_localbasis_cache(element)
    lfs = name_leaf_lfs(element, restriction)

    def decl(name, shape, shape_impl):
        return "typename {}::{}ReturnType {};".format(t_cache,
                                                      which,
                                                      name,
                                                      )
    return decl


@backend(interface="evaluate_basis")
@kernel_cached
def evaluate_basis(leaf_element, name, restriction):
    lfs = name_leaf_lfs(leaf_element, restriction)
    temporary_variable(name, shape=(name_lfs_bound(lfs),), decl_method=declare_cache_temporary(leaf_element, restriction, 'Function'))
    cache = name_localbasis_cache(leaf_element)
    qp = get_backend("qp_in_cell")(restriction)
    instruction(inames=get_backend("quad_inames")(),
                code='{} = {}.evaluateFunction({}, {}.finiteElement().localBasis());'.format(name,
                                                                                             cache,
                                                                                             str(qp),
                                                                                             lfs,
                                                                                             ),
                assignees=frozenset({name}),
                read_variables=frozenset({get_pymbolic_basename(qp)}),
                )


def pymbolic_basis(leaf_element, restriction, number, context=''):
    assert leaf_element.num_sub_elements() == 0
    name = "phi_{}".format(FEM_name_mangling(leaf_element))
    name = restricted_name(name, restriction)
    evaluate_basis(leaf_element, name, restriction)
    iname, = lfs_inames(leaf_element, restriction, number, context=context)

    return Subscript(Variable(name), (Variable(iname),))


@backend(interface="evaluate_grad")
@kernel_cached
def evaluate_reference_gradient(leaf_element, name, restriction):
    lfs = name_leaf_lfs(leaf_element, restriction)
    temporary_variable(name, shape=(name_lfs_bound(lfs), 1, name_dimension()), decl_method=declare_cache_temporary(leaf_element, restriction, 'Jacobian'))
    cache = name_localbasis_cache(leaf_element)
    qp = get_backend("qp_in_cell")(restriction)
    instruction(inames=get_backend("quad_inames")(),
                code='{} = {}.evaluateJacobian({}, {}.finiteElement().localBasis());'.format(name,
                                                                                             cache,
                                                                                             str(qp),
                                                                                             lfs,
                                                                                             ),
                assignees=frozenset({name}),
                read_variables=frozenset({get_pymbolic_basename(qp)}),
                )


def pymbolic_reference_gradient(leaf_element, restriction, number, context=''):
    assert leaf_element.num_sub_elements() == 0
    name = "js_{}".format(FEM_name_mangling(leaf_element))
    name = restricted_name(name, restriction)
    evaluate_reference_gradient(leaf_element, name, restriction)
    iname, = lfs_inames(leaf_element, restriction, number, context=context)

    return Subscript(Variable(name), (Variable(iname), 0))


def shape_as_pymbolic(shape):
    def _shape_as_pymbolic(s):
        if isinstance(s, str):
            return Variable(s)
        else:
            return s
    return tuple(_shape_as_pymbolic(s) for s in shape)


@kernel_cached
def evaluate_coefficient(element, name, container, restriction, component):
    from ufl.functionview import select_subelement
    sub_element = select_subelement(element, component)

    # Determine the rank of the trialfunction tensor
    rank = len(sub_element.value_shape())

    shape = sub_element.value_shape()
    shape_impl = ('arr', ) * rank
    idims = tuple(dimension_iname(count=i) for i in range(rank))
    leaf_element = sub_element
    from ufl import VectorElement, TensorElement
    if isinstance(sub_element, (VectorElement, TensorElement)):
        leaf_element = sub_element.sub_elements()[0]

    temporary_variable(name, shape=shape, shape_impl=shape_impl)
    lfs = name_lfs(element, restriction, component)
    basis = pymbolic_basis(leaf_element, restriction, 0, context='trial')
    index, = get_pymbolic_indices(basis)

    if isinstance(sub_element, (VectorElement, TensorElement)):
        lfs = lfs_child(lfs, idims, shape=shape_as_pymbolic(shape), symmetry=element.symmetry())

    from dune.perftool.pdelab.argument import pymbolic_coefficient
    coeff = pymbolic_coefficient(container, lfs, index)
    assignee = Subscript(Variable(name), tuple(Variable(i) for i in idims))

    reduction_expr = Product((coeff, basis))
    instruction(expression=Reduction("sum", index, reduction_expr, allow_simultaneous=True),
                assignee=assignee,
                forced_iname_deps=frozenset(get_backend("quad_inames")()).union(frozenset(idims)),
                forced_iname_deps_is_final=True,
                )


@kernel_cached
def evaluate_coefficient_gradient(element, name, container, restriction, component):
    # First we determine the rank of the tensor we are talking about
    from ufl.functionview import select_subelement
    sub_element = select_subelement(element, component)
    rank = len(sub_element.value_shape()) + 1

    # We do then set some variables accordingly
    shape = sub_element.value_shape() + (element.cell().geometric_dimension(),)
    shape_impl = ('arr',) * rank

    idims = tuple(dimension_iname(count=i) for i in range(rank))
    leaf_element = sub_element
    from ufl import VectorElement, TensorElement
    if isinstance(sub_element, (VectorElement, TensorElement)):
        leaf_element = sub_element.sub_elements()[0]

    # and proceed to call the necessary generator functions
    temporary_variable(name, shape=shape, shape_impl=shape_impl)
    lfs = name_lfs(element, restriction, component)
    basis = pymbolic_reference_gradient(leaf_element, restriction, 0, context='trialgrad')
    index, _ = get_pymbolic_indices(basis)

    if isinstance(sub_element, (VectorElement, TensorElement)):
        lfs = lfs_child(lfs, idims[:-1], shape=shape_as_pymbolic(shape[:-1]), symmetry=element.symmetry())

    from dune.perftool.pdelab.argument import pymbolic_coefficient
    coeff = pymbolic_coefficient(container, lfs, index)
    assignee = Subscript(Variable(name), tuple(Variable(i) for i in idims))

    reduction_expr = Product((coeff, Subscript(Variable(get_pymbolic_basename(basis)), basis.index + (Variable(idims[-1]),))))

    instruction(expression=Reduction("sum", index, reduction_expr, allow_simultaneous=True),
                assignee=assignee,
                forced_iname_deps=frozenset(get_backend("quad_inames")()).union(frozenset(idims)),
                forced_iname_deps_is_final=True,
                )
