import copy

from dune.perftool.loopy.symbolic import substitute
from dune.perftool.pdelab.argument import (name_accumulation_variable,
                                           name_coefficientcontainer,
                                           pymbolic_coefficient,
                                           PDELabAccumulationFunction,
                                           )
from dune.perftool.generation import (backend,
                                      barrier,
                                      built_instruction,
                                      domain,
                                      function_mangler,
                                      get_counter,
                                      get_global_context_value,
                                      globalarg,
                                      iname,
                                      instruction,
                                      retrieve_cache_items,
                                      silenced_warning,
                                      temporary_variable,
                                      transform,
                                      )
from dune.perftool.options import get_option
from dune.perftool.loopy.flatten import flatten_index
from dune.perftool.loopy.buffer import (get_buffer_temporary,
                                        initialize_buffer,
                                        switch_base_storage,
                                        )
from dune.perftool.sumfact.quadrature import nest_quadrature_loops
from dune.perftool.pdelab.localoperator import determine_accumulation_space
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.pdelab.spaces import (name_lfs,
                                         name_lfs_bound,
                                         )
from dune.perftool.sumfact.amatrix import (AMatrix,
                                           LargeAMatrix,
                                           quadrature_points_per_direction,
                                           basis_functions_per_direction,
                                           name_large_theta,
                                           name_theta,
                                           )
from dune.perftool.loopy.symbolic import SumfactKernel
from dune.perftool.tools import get_pymbolic_basename
from dune.perftool.error import PerftoolError
from pymbolic.primitives import (Call,
                                 Product,
                                 Subscript,
                                 Variable,
                                 )
from dune.perftool.sumfact.quadrature import quadrature_inames
from dune.perftool.sumfact.vectorization import find_sumfact
from loopy import Reduction, GlobalArg
from loopy.symbolic import FunctionIdentifier, IdentityMapper

import loopy as lp
import numpy as np
import pymbolic.primitives as prim
from pytools import product


class IndexFiddleMapper(IdentityMapper):
    def __init__(self, var, index, pos):
        assert isinstance(var, str)
        self.var = var
        self.index = index
        if isinstance(index, str):
            self.index = prim.Variable(index)
        self.pos = pos

    def map_variable(self, expr):
        if expr.name == self.var:
            return prim.Subscript(expr, (self.var,))
        else:
            return IdentityMapper.map_variable(self, expr)

    def map_subscript(self, expr):
        if expr.aggregate.name == self.var:
            ind = expr.index
            if not isinstance(ind, tuple):
                ind = (ind,)
            if self.pos is None:
                ind = ind + (self.index,)
            else:
                raise NotImplementedError
            return prim.Subscript(expr.aggregate, ind)
        else:
            return IdentityMapper.map_subscript(self, expr)


def fiddle_in_index(expr, var, index, pos=None):
    return IndexFiddleMapper(var, index, pos)(expr)


def default_resolution(insns):
    for insn in insns:
        if isinstance(insn, (lp.Assignment, lp.CallInstruction)):
            replace = {}
            deps = []
            for sumf in find_sumfact(insn.expression):
                # Maybe set up the input
                if sumf.setup_method:
                    shape = product(mat.cols for mat in sumf.a_matrices)
                    shape = (shape,)
                    inp = get_buffer_temporary(sumf.buffer, shape=shape)
                    silenced_warning('read_no_write({})'.format(inp))

                    func, args = sumf.setup_method
                    insn_dep = frozenset({func(inp, *args)})
                else:
                    insn_dep = sumf.insn_dep

                # Call the sum factorization algorithm
                var, dep = sum_factorization_kernel(sumf.a_matrices, sumf.buffer, insn_dep, sumf.additional_inames)
                replace[sumf] = prim.Variable(var)
                deps.append(dep)

            if replace:
                built_instruction(insn.copy(expression=substitute(insn.expression, replace),
                                            depends_on=frozenset(*deps)
                                            )
                                  )
                if isinstance(insn, lp.Assignment):
                    from dune.perftool.generation.loopy import expr_instruction_impl
                    expr_instruction_impl.remove_by_value(insn)
                if isinstance(insn, lp.CallInstruction):
                    from dune.perftool.generation.loopy import call_instruction_impl
                    call_instruction_impl.remove_by_value(insn)


def apply_sumfact_grad_vectorization(insns, stage):
    sumfact_kernels = []
    for insn in insns:
        if isinstance(insn, (lp.Assignment, lp.CallInstruction)):
            sumf = find_sumfact(insn.expression)
            if sumf:
                sumf, = sumf
                if sumf.stage == stage:
                    sumfact_kernels.append(sumf)

    # Now apply some heuristics when to vectorize...
    if len(set(sumfact_kernels)) < 3:
        return
    else:
        # Vectorize!!!
        sumfact_kernels = sorted(sumfact_kernels, key=lambda s: s.preferred_interleaving_position)
        # Pad to 4
        if len(sumfact_kernels) < 4:
            sumfact_kernels.append(sumfact_kernels[0])

        # Determine a name for the buffer in use
        buffer = "_".join(sf.buffer for sf in sumfact_kernels)
        insn_dep = frozenset()

        # Maybe initialize the input buffer
        if sumfact_kernels[0].setup_method:
            assert stage == 1
            shape = product(mat.cols for mat in sumfact_kernels[0].a_matrices)
            shape = (shape, 4)
            initialize_buffer(buffer, base_storage_size=4 * shape[0])
            inp = get_buffer_temporary(buffer, shape=shape)
            silenced_warning('read_no_write({})'.format(inp))

            # Call the individual setup_methods
            for i, sumf in enumerate(sumfact_kernels):
                func, args = sumf.setup_method
                insn_dep = insn_dep.union({func(inp, *args, additional_indices=(i,))})
        else:
            assert stage == 3
            # No setup method defined. We need to make sure the input is correctly setup
            shape = tuple(mat.cols for mat in sumfact_kernels[0].a_matrices) + (4,)
            initialize_buffer(buffer, base_storage_size=4 * shape[0])
            dim_tags = ",".join("f" * (len(shape) - 1)) + ",c"
            inp = get_buffer_temporary(buffer, shape=shape, dim_tags=dim_tags)

            for i, sumf in enumerate(sumfact_kernels):
                assert sumf.input_temporary
                for insn in insns:
                    if isinstance(insn, lp.Assignment):
                        if get_pymbolic_basename(insn.assignee) == sumf.input_temporary:
                            built_instruction(insn.copy(assignee=prim.Subscript(prim.Variable(inp), insn.assignee.index + (i,)),
                                                        id=insn.id + "__{}".format(i)))
                            insn_dep = insn_dep.union(insn.depends_on)
                            from dune.perftool.generation.loopy import expr_instruction_impl
                            expr_instruction_impl.remove_by_value(insn)

        # Determine the joined AMatrix
        large_a_matrices = []
        for i in range(len(sumfact_kernels[0].a_matrices)):
            assert len(set(tuple(sf.a_matrices[i].rows for sf in sumfact_kernels))) == 1
            assert len(set(tuple(sf.a_matrices[i].cols for sf in sumfact_kernels))) == 1
            large = LargeAMatrix(rows=sumfact_kernels[0].a_matrices[i].rows,
                                 cols=sumfact_kernels[0].a_matrices[i].cols,
                                 transpose=tuple(sf.a_matrices[i].transpose for sf in sumfact_kernels),
                                 derivative=tuple(sf.a_matrices[i].derivative for sf in sumfact_kernels),
                                 )
            large_a_matrices.append(large)

        # Join the instruction dependencies
        insn_dep = insn_dep.union(*tuple(sf.insn_dep for sf in sumfact_kernels))

        var, dep = sum_factorization_kernel(large_a_matrices, buffer, insn_dep, sumfact_kernels[0].additional_inames, add_vec_tag=True)

        for insn in insns:
            if isinstance(insn, (lp.Assignment, lp.CallInstruction)):
                sumfacts = find_sumfact(insn.expression)
                if sumfacts:
                    sumf, = sumfacts
                    if sumf.stage == stage:
                        replace = {}
                        replace[sumf] = prim.Variable(var)
                        newexpr = substitute(insn.expression, replace)
                        newexpr = fiddle_in_index(newexpr, var, sumf.preferred_interleaving_position)
                        assert newexpr
                        built_instruction(insn.copy(expression=newexpr,
                                                    depends_on=dep,
                                                    ))

                        if isinstance(insn, lp.Assignment):
                            from dune.perftool.generation.loopy import expr_instruction_impl
                            expr_instruction_impl.remove_by_value(insn)
                        if isinstance(insn, lp.CallInstruction):
                            from dune.perftool.generation.loopy import call_instruction_impl
                            call_instruction_impl.remove_by_value(insn)


def expand_sumfact_kernels(tag):
    if get_option("vectorize_grads"):
        apply_sumfact_grad_vectorization([i for i in retrieve_cache_items("{} and instruction".format(tag))], 1)
        apply_sumfact_grad_vectorization([i for i in retrieve_cache_items("{} and instruction".format(tag))], 3)
    default_resolution(retrieve_cache_items("{} and instruction".format(tag)))


@iname
def _sumfact_iname(bound, _type, count):
    name = "sf_{}_{}".format(_type, str(count))
    domain(name, bound)
    return name


def sumfact_iname(bound, _type):
    count = get_counter('_sumfac_iname_{}'.format(_type))
    name = _sumfact_iname(bound, _type, count)
    return name


def setup_theta(inp, element, restriction, component, index):
    if index is None:
        index = ()
    else:
        index = (index,)
    # Write initial coefficients into buffer
    lfs = name_lfs(element, restriction, component)
    basisiname = sumfact_iname(name_lfs_bound(lfs), "basis")
    container = name_coefficientcontainer(restriction)
    coeff = pymbolic_coefficient(container, lfs, basisiname)
    assignee = Subscript(Variable(inp), (Variable(basisiname),) + index)
    return instruction(assignee=assignee,
                       expression=coeff,
                       )


def name_test_function_contribution(test):
    count = get_counter('sumfact_contribution_counter')
    grad = "grad_" if test.reference_grad else ""
    return restricted_name("contrib_{}phi_{}".format(grad, str(count)), test.restriction)


@backend(interface="accum_insn", name="sumfact")
def generate_accumulation_instruction(visitor, accterm, measure, subdomain_id):
    pymbolic_expr = visitor(accterm.term)

    if pymbolic_expr == 0:
        return

    # Get geometric dimension
    formdata = get_global_context_value('formdata')
    dim = formdata.geometric_dimension

    # Collect buffers we need
    buffers = []

    # If this is a gradient, we find the gradient iname
    additional_inames = frozenset({})
    if accterm.argument.index:
        for i in accterm.argument.index._indices:
            if i not in visitor.dimension_indices:
                from dune.perftool.pdelab.localoperator import grad_iname
                additional_inames = additional_inames.union(frozenset({grad_iname(i, dim)}))
                for i in range(dim):
                    buffers.append(name_test_function_contribution(accterm.argument))
    else:
        buffers.append(name_test_function_contribution(accterm.argument))

    # Collect a list of instruction IDs of the contributions
    contribution_ids = []

    # TODO covers only 2D
    for i, buf in enumerate(buffers):
        # Get the a matrices needed for this accumulation term
        rows = basis_functions_per_direction()
        cols = quadrature_points_per_direction()
        theta_matrix = AMatrix(rows, cols, transpose=True)

        # If this is a gradient we need different matrices
        if accterm.argument.index:
            rows = basis_functions_per_direction()
            cols = quadrature_points_per_direction()
            dtheta_matrix = AMatrix(rows, cols, transpose=True, derivative=True)

            a_matrices = [theta_matrix] * dim
            a_matrices[i] = dtheta_matrix
            a_matrices = tuple(a_matrices)
            pref_pos = i
        else:
            a_matrices = (theta_matrix,) * dim
            pref_pos = None

        # Get the vectorization info. If this happens during the dry run, we get dummies
        from dune.perftool.sumfact.vectorization import get_vectorization_info
        a_matrices, buffer, input, index = get_vectorization_info(a_matrices)

        # Initialize a base storage for this buffer and get a temporay pointing to it
#        shape = product(mat.cols for mat in a_matrices)
#        if vec:
#            shape = (shape, 4)
        temp = initialize_buffer(buffer,
                                 base_storage_size=product(max(mat.rows, mat.cols) for mat in a_matrices),
                                 num=2
                                 ).get_temporary(shape=(quadrature_points_per_direction(),) * dim,
                                                 dim_tags=",".join(['f'] * dim),
                                                 name=input,
                                                 )

        # Replace gradient iname with correct index for assignement
        replace_dict = {}
        for iname in additional_inames:
            replace_dict[Variable(iname)] = i
        expression = substitute(pymbolic_expr, replace_dict)

        # Determine dependencies
        from loopy.match import Or, Writes
        from pymbolic.mapper.dependency import DependencyMapper
        from dune.perftool.tools import get_pymbolic_basename
        deps = Or(tuple(Writes(get_pymbolic_basename(expr)) for expr in DependencyMapper()(expression)))

        # Issue an instruction in the quadrature loop that fills the buffer
        # with the evaluation of the contribution at all quadrature points
        assignee = Subscript(Variable(temp), tuple(Variable(i) for i in quadrature_inames()))
        contrib_dep = instruction(assignee=assignee,
                                  expression=expression,
                                  forced_iname_deps=frozenset(quadrature_inames() + visitor.inames),
                                  forced_iname_deps_is_final=True,
                                  tags=frozenset({"quadvec"}),
                                  depends_on=frozenset({deps})
                                  )
        contribution_ids.append(contrib_dep)

        # Add a sum factorization kernel that implements the multiplication
        # with the test function (stage 3)
        result, insn_dep = sum_factorization_kernel(a_matrices,
                                                    buffer,
                                                    3,
                                                    insn_dep=frozenset({contrib_dep}),
                                                    additional_inames=frozenset(visitor.inames),
                                                    preferred_position=pref_pos,
                                                    )

        inames = tuple(sumfact_iname(mat.rows, 'accum') for mat in a_matrices)

        # Collect the lfs and lfs indices for the accumulate call
        test_lfs = determine_accumulation_space(accterm.argument.expr, 0, measure)
        test_lfs.index = flatten_index(tuple(Variable(i) for i in inames),
                                       (basis_functions_per_direction(),) * dim,
                                       order="f"
                                       )

        # In the jacobian case, also determine the space for the ansatz space
        ansatz_lfs = determine_accumulation_space(accterm.term, 1, measure)
        rank = 2 if visitor.inames else 1
        if rank == 2:
            ansatz_lfs.index = flatten_index(tuple(Variable(i) for i in visitor.inames),
                                             (basis_functions_per_direction(),) * dim,
                                             order="f"
                                             )

        # Construct the expression representing "{r,jac}.accumulate(..)"
        accum = name_accumulation_variable()
        expr = Call(PDELabAccumulationFunction(accum, rank),
                    (ansatz_lfs.get_args() +
                     test_lfs.get_args() +
                     (Subscript(result, tuple(Variable(i) for i in inames)),)
                     )
                    )
        instruction(assignees=(),
                    expression=expr,
                    forced_iname_deps=frozenset(inames + visitor.inames),
                    forced_iname_deps_is_final=True,
                    depends_on=insn_dep,
                    )

        # Mark the transformation that moves the quadrature loop inside the trialfunction loops for application
        transform(nest_quadrature_loops, visitor.inames)


def sum_factorization_kernel(a_matrices, buf, stage, insn_dep=frozenset({}), additional_inames=frozenset({}), add_vec_tag=False, preferred_position=None):
    """
    Calculate a sum factorization matrix product.

    Y = A_{d-1}*...*A_0*X

    where X is the input matrix and Y is the output variable.

    Arguments:
    ----------
    a_matrices: An iterable of AMatrix instances
        The list of tensors to be applied to the input.
        Order of application is from 0 up.
    buf: A string identifying the flip flop buffer in use
        for intermediate results. The memory is expected to be
        pre-initialized with the input.
    insn_dep: an instruction ID that the first issued instruction
        should depend upon. All following ones will depend on each
        other.
    """
    if get_global_context_value("dry_run", False):
        return SumfactKernel(a_matrices, buf, stage, preferred_position), frozenset()

    ftags = "f,f" + (",vec" if add_vec_tag else "")
    ctags = "c,c" + (",vec" if add_vec_tag else "")
    vec_shape = (4,) if add_vec_tag else ()

    insn_dep = frozenset({barrier(depends_on=insn_dep,
                                  within_inames=additional_inames,
                                  )})

    for l, a_matrix in enumerate(a_matrices):
        # Get a temporary that interprets the base storage of the input
        # as a column-major matrix. In later iteration of the amatrix loop
        # this reinterprets the output of the previous iteration.
        inp_shape = (a_matrix.cols, product(mat.rows for mat in a_matrices[:l]) * product(mat.cols for mat in a_matrices[l + 1:]))
        inp = get_buffer_temporary(buf,
                                   shape=inp_shape + vec_shape,
                                   dim_tags=ftags)

        # The input temporary will only be read from, so we need to silence the loopy warning
        silenced_warning('read_no_write({})'.format(inp))

        switch_base_storage(buf)

        # Get a temporary that interprets the base storage of the output
        # as row-major matrix
        out_shape = (a_matrix.rows, product(mat.rows for mat in a_matrices[:l]) * product(mat.cols for mat in a_matrices[l + 1:]))

        out = get_buffer_temporary(buf,
                                   shape=out_shape + vec_shape,
                                   dim_tags=ctags)

        # Get the inames needed for one matrix-matrix multiplication
        i = sumfact_iname(out_shape[0], "row")
        j = sumfact_iname(out_shape[1], "col")
        k = sumfact_iname(a_matrix.cols, "red")

        # Select the correct Theta Matrix for multiplication
        if isinstance(a_matrix, AMatrix):
            theta = name_theta(transpose=a_matrix.transpose, derivative=a_matrix.derivative)
            vec_iname = ()
        elif isinstance(a_matrix, LargeAMatrix):
            theta = name_large_theta(transpose=a_matrix.transpose, derivative=a_matrix.derivative)
            iname = sumfact_iname(4, "vec")
            vec_iname = (prim.Variable(iname),)
            transform(lp.tag_inames, [(iname, "vec")])
        else:
            raise TypeError("What kind of an AMatrix is that?")

        # Construct the matrix-matrix-multiplication expression a_ik*in_kj
        prod = Product((Subscript(Variable(theta), (Variable(i), Variable(k)) + vec_iname),
                        Subscript(Variable(inp), (Variable(k), Variable(j)) + vec_iname)
                        ))

        # Issue the reduction instruction that implements the multiplication
        # at the same time store the instruction ID for the next instruction to depend on
        insn_dep = frozenset({instruction(assignee=Subscript(Variable(out), (Variable(i), Variable(j)) + vec_iname),
                                          expression=Reduction("sum", k, prod),
                                          forced_iname_deps=frozenset({i, j}).union(additional_inames),
                                          forced_iname_deps_is_final=True,
                                          depends_on=insn_dep,
                                          )
                              })

    # Get geometric dimension
    formdata = get_global_context_value('formdata')
    dim = formdata.geometric_dimension

    out_shape = tuple(mat.rows for mat in a_matrices)
    dim_tags = ",".join(['f'] * dim)

    if add_vec_tag:
        out_shape = out_shape + vec_shape
        dim_tags = dim_tags + ",c"

    out = get_buffer_temporary(buf,
                               shape=out_shape,
                               dim_tags=dim_tags)
    silenced_warning('read_no_write({})'.format(out))

    return prim.Variable(out), insn_dep
